# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.


class FunctionExtensionException(Exception):
    """Excpetion emitted from Azure Functions Python Worker extension
    """
    pass
